# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Performs float casting quantization."""

from typing import Any, Optional

import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.utils import common_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils

ALGORITHM_KEY = "float_casting"
_TFLOpName = qtyping.TFLOperationName
_QuantTransformation = qtyping.QuantTransformation

# "Fake" quantization config for float casting.
_FP16_QUANT_CONFIG = qtyping.TensorQuantizationConfig(
    num_bits=16, dtype=qtyping.TensorDataType.FLOAT
)


# Ops that support weight quantization config (e.g., support Weight-only).
SUPPORTED_WEIGHT_QUANT_OPS = frozenset([
    _TFLOpName.FULLY_CONNECTED,
    _TFLOpName.CONV_2D,
    _TFLOpName.DEPTHWISE_CONV_2D,
    _TFLOpName.CONV_2D_TRANSPOSE,
    _TFLOpName.EMBEDDING_LOOKUP,
])


def check_op_quantization_config(
    op_name: _TFLOpName,
    op_quant_config: qtyping.OpQuantizationConfig,
    config_check_policy: Optional[qtyping.ConfigCheckPolicyDict] = None,
) -> None:
  """Checks if the op is valid for float casting quantization.

  Args:
    op_name: The name of the op.
    op_quant_config: The quantization config for the op.
    config_check_policy: The policy to check the quantization config.

  Raises:
    ValueError: If the op is not supported or the compute_precision is not
      FLOAT.
  """
  # TODO: b/353780772 - Add config check policy for float casting quantization.
  if config_check_policy is not None and config_check_policy:
    raise ValueError(f"Config check isn't implemented yet for op: {op_name}.")

  # Check if WEIGHT_ONLY.
  if op_quant_config.compute_precision != qtyping.ComputePrecision.FLOAT:
    raise ValueError(
        "Currently, only Weight-Only is supported for float casting"
        " quantization. Got unsupported execution mode:"
        f" {op_quant_config.compute_precision} for op: {op_name}"
    )
  if op_quant_config.activation_tensor_config is not None:
    raise ValueError(
        "Activation tensor quantization is not supported for float casting"
        " quantization."
    )
  if op_name not in SUPPORTED_WEIGHT_QUANT_OPS:
    raise ValueError(
        f"Unsupported op: {op_name} for float casting quantization."
    )
  if op_quant_config.weight_tensor_config is None:
    raise ValueError(
        "Weight tensor quantization config is required for float casting"
        " quantization."
    )
  if (
      op_quant_config.weight_tensor_config.num_bits != 16
      or op_quant_config.weight_tensor_config.dtype
      != qtyping.TensorDataType.FLOAT
  ):
    raise ValueError(
        "Currently, float casting quantization config requires number of bits"
        " to be set as 16, dtype as float, got"
        f" {op_quant_config.weight_tensor_config.num_bits} and"
        f" {op_quant_config.weight_tensor_config.dtype} ."
    )


def materialize_fc_conv(
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],  # pylint: disable=unused-argument
    tensor_quant_params_cache: common_utils.TensorQuantParamsCache,
) -> list[qtyping.TensorTransformationParams]:
  """Materialize tensors in fully_connected, conv_2d and depthwise_conv_2d ops.

  This function is called by the quantization pipeline to materialize
  quantization parameters for the weight tensor of the op.

  Args:
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    tensor_name_to_qsv: A map of tensor name to quantization parameters
      (unused).
    tensor_quant_params_cache: Cache of already computed
      `UniformQuantParams|NonLinearQuantParams` objects keyed on a tuple of the
      buffer ID and the `TensorQuantizationConfig` used to compute it.

  Returns:
    Quantization configuration for the weight tensor of the op.

  Raises:
    ValueError: If the op is not supported or the compute precision is not
      FLOAT.
  """

  input_tensor, weight_tensor, bias_tensor, output_tensor = (
      tfl_flatbuffer_utils.parse_fc_bmm_conv_tensors(
          op_info.op, graph_info.subgraph_tensors
      )
  )
  op_tensor_params = []
  # Input tensor.
  input_quant_params = _config_no_quantize_tensor(
      op_info, input_tensor, is_inbounding_tensor=True
  )
  op_tensor_params.append(input_quant_params)
  # Weight tensor.
  weight_content = tfl_flatbuffer_utils.get_tensor_data(
      weight_tensor,
      graph_info.buffers,
  )
  assert weight_content is not None
  if not (
      quant_params := tensor_quant_params_cache.lookup(
          weight_tensor.buffer, _FP16_QUANT_CONFIG
      )
  ):
    quant_params = qtyping.NonLinearQuantParams(
        num_bits=16, quantized_data=weight_content.astype(np.float16)
    )
    tensor_quant_params_cache.insert(
        weight_tensor.buffer, _FP16_QUANT_CONFIG, quant_params
    )
  op2weight_params = qtyping.OpToTensorParams(
      subgraph_op_id=op_info.subgraph_op_index,
      parameters=quant_params,
      transformations=[_QuantTransformation.ADD_DEQUANTIZE],
  )
  op_tensor_params.append(
      qtyping.TensorTransformationParams(
          tensor_name=tfl_flatbuffer_utils.get_tensor_name(weight_tensor),
          consumers=[op2weight_params],
      )
  )
  # Output tensor.
  output_quant_params = _config_no_quantize_tensor(
      op_info, output_tensor, is_inbounding_tensor=False
  )
  op_tensor_params.append(output_quant_params)
  # Bias tensor.
  if bias_tensor is not None:
    bias_quant_params = _config_no_quantize_tensor(
        op_info, bias_tensor, is_inbounding_tensor=True
    )
    op_tensor_params.append(bias_quant_params)
  return op_tensor_params


def materialize_embedding_lookup(
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: common_utils.TensorQuantParamsCache,
) -> list[qtyping.TensorTransformationParams]:
  return materialize_fc_conv(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_conv2d_transpose(
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],  # pylint: disable=unused-argument
    tensor_quant_params_cache: common_utils.TensorQuantParamsCache,
) -> list[qtyping.TensorTransformationParams]:
  """Materialize tensors in fully_connected, conv_2d and depthwise_conv_2d ops.

  This function is called by the quantization pipeline to materialize
  quantization parameters for the weight tensor of the op.

  Args:
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    tensor_name_to_qsv: A map of tensor name to quantization parameters
      (unused).
    tensor_quant_params_cache: Cache of already computed
      `UniformQuantParams|NonLinearQuantParams` objects keyed on a tuple of the
      buffer ID and the `TensorQuantizationConfig` used to compute it.

  Returns:
    Quantization configuration for the weight tensor of the op.

  Raises:
    ValueError: If the op is not supported or the execution mode is not
      WEIGHT_ONLY.
  """

  input_tensor, weight_tensor, bias_tensor, output_tensor = (
      tfl_flatbuffer_utils.parse_fc_bmm_conv_tensors(
          op_info.op,
          graph_info.subgraph_tensors,
          input_index=2,
          weight_index=1,
          bias_index=3,
          output_index=0,
      )
  )
  op_tensor_params = []
  # Input tensor.
  input_quant_params = _config_no_quantize_tensor(
      op_info, input_tensor, is_inbounding_tensor=True
  )
  op_tensor_params.append(input_quant_params)
  # Weight tensor.
  weight_content = tfl_flatbuffer_utils.get_tensor_data(
      weight_tensor,
      graph_info.buffers,
  )
  assert weight_content is not None
  if not (
      quant_params := tensor_quant_params_cache.lookup(
          weight_tensor.buffer, _FP16_QUANT_CONFIG
      )
  ):
    quant_params = qtyping.NonLinearQuantParams(
        num_bits=16, quantized_data=weight_content.astype(np.float16)
    )
    tensor_quant_params_cache.insert(
        weight_tensor.buffer, _FP16_QUANT_CONFIG, quant_params
    )
  op2weight_params = qtyping.OpToTensorParams(
      subgraph_op_id=op_info.subgraph_op_index,
      parameters=quant_params,
      transformations=[_QuantTransformation.ADD_DEQUANTIZE],
  )
  op_tensor_params.append(
      qtyping.TensorTransformationParams(
          tensor_name=tfl_flatbuffer_utils.get_tensor_name(weight_tensor),
          consumers=[op2weight_params],
      )
  )
  # Output tensor.
  output_quant_params = _config_no_quantize_tensor(
      op_info, output_tensor, is_inbounding_tensor=False
  )
  op_tensor_params.append(output_quant_params)
  # Bias tensor.
  if bias_tensor is not None:
    bias_quant_params = _config_no_quantize_tensor(
        op_info, bias_tensor, is_inbounding_tensor=True
    )
    op_tensor_params.append(bias_quant_params)
  return op_tensor_params


def _config_no_quantize_tensor(
    op_info: qtyping.OpInfo,
    tensor: Any,
    is_inbounding_tensor: bool,
) -> qtyping.TensorTransformationParams:
  """Configures a tensor to be not quantized.

  Args:
    op_info: Aggregated information about the op (e.g., quantization config).
    tensor: The tensor to be configured.
    is_inbounding_tensor: Whether the tensor is an inbounding tensor.

  Returns:
    TensorTransformationParams for the tensor.
  """
  tensor_name = tfl_flatbuffer_utils.get_tensor_name(tensor)
  op2tensor_params = qtyping.OpToTensorParams(
      subgraph_op_id=op_info.subgraph_op_index,
      transformations=[_QuantTransformation.NO_QUANTIZE],
  )
  if is_inbounding_tensor:
    return qtyping.TensorTransformationParams(
        tensor_name=tensor_name,
        consumers=[op2tensor_params],
    )
  return qtyping.TensorTransformationParams(
      tensor_name=tensor_name, producer=op2tensor_params
  )


def init_qsvs(*_) -> qtyping.QSV:
  """Currently calibration free. Placeholder for AlgorithmManager."""
  return {}


def calibrate(*_) -> dict[str, qtyping.QSV]:
  """Currently calibration free. Placeholder for AlgorithmManager."""
  return {}



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import io
import sys
import time
import tracemalloc
from unittest import mock

import absl.testing.absltest as absltest
from ai_edge_quantizer.utils import progress_utils


class ProgressBarTest(absltest.TestCase):

  @mock.patch('tqdm.tqdm')
  def test_progress_bar_update(self, mock_tqdm):
    mock_progress_bar_instance = mock_tqdm.return_value
    with progress_utils.ProgressBar(total_steps=10) as pb:
      pb.update_single_step()
      pb.update_single_step()

    mock_tqdm.assert_called_once_with(
        total=10, desc='', leave=True, disable=False
    )
    self.assertEqual(mock_progress_bar_instance.update.call_count, 2)
    mock_progress_bar_instance.update.assert_called_with(1)
    mock_progress_bar_instance.close.assert_called_once()

  @mock.patch('tqdm.tqdm')
  def test_progress_bar_disable(self, mock_tqdm):
    mock_progress_bar_instance = mock_tqdm.return_value
    with progress_utils.ProgressBar(total_steps=10, disable=True):
      pass
    mock_tqdm.assert_called_once_with(
        total=10, desc='', leave=True, disable=True
    )
    mock_progress_bar_instance.close.assert_called_once()

  @mock.patch('tqdm.tqdm')
  def test_progress_bar_disappear_on_finish(self, mock_tqdm):
    mock_progress_bar_instance = mock_tqdm.return_value
    with progress_utils.ProgressBar(total_steps=10, disappear_on_finish=True):
      pass
    mock_tqdm.assert_called_once_with(
        total=10, desc='', leave=False, disable=False
    )
    mock_progress_bar_instance.close.assert_called_once()


class ProgressReportTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.mock_time = self.enter_context(
        mock.patch.object(time, 'time', autospec=True)
    )
    self.mock_tracemalloc_start = self.enter_context(
        mock.patch.object(tracemalloc, 'start', autospec=True)
    )
    self.mock_tracemalloc_get_traced_memory = self.enter_context(
        mock.patch.object(tracemalloc, 'get_traced_memory', autospec=True)
    )

  def test_generate_progress_report(self):
    self.mock_time.side_effect = [100.0, 105.5]  # Start time, end time.
    # Mock memory: current=1MB, peak=2MB.
    self.mock_tracemalloc_get_traced_memory.return_value = (
        1 * 1024 * 1024,
        2 * 1024 * 1024,
    )

    progress_report = progress_utils.ProgressReport()
    progress_report.capture_progess_start()

    original_model = b'\x01' * 2048  # 2KB.
    quantized_model = b'\x02' * 1024  # 1KB

    mock_stdout = io.StringIO()
    with mock.patch.object(sys, 'stdout', mock_stdout):
      progress_report.generate_progress_report(original_model, quantized_model)

    self.mock_tracemalloc_start.assert_called_once()
    output = mock_stdout.getvalue()
    self.assertIn('Original model size: 2.00 KB', output)
    self.assertIn('Quantized model size: 1.00 KB', output)
    self.assertIn('Quantization Ratio: 0.50', output)
    self.assertIn('Total time: 5.50 seconds', output)
    self.assertIn('Memory peak: 2.00 MB', output)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""E2E tests for the quantizer for model input and output."""

from absl.testing import parameterized
import absl.testing.absltest as absltest
import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer import quantizer
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils


_OpExecutionMode = qtyping.OpExecutionMode
_OpName = qtyping.TFLOperationName
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_Granularity = qtyping.QuantGranularity
_OpQuantConfig = qtyping.OpQuantizationConfig
_ComputePrecision = qtyping.ComputePrecision

_RNG = np.random.default_rng(66)


def _get_dummy_data(num_samples):
  samples = []
  for _ in range(num_samples):
    samples.append(
        {'input_1': _RNG.uniform(size=(1, 32, 32)).astype(np.float32)}
    )

  data = {
      tfl_interpreter_utils.DEFAULT_SIGNATURE_KEY: samples,
  }
  return data


def _get_calibration_data(num_samples: int = 128):
  return _get_dummy_data(num_samples)


def _get_test_data(num_samples: int = 8):
  return _get_dummy_data(num_samples)


class InputOutputTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.float_model_path = test_utils.get_path_to_datafile(
        '../models/single_mean.tflite'
    )
    self._quantizer = quantizer.Quantizer(self.float_model_path)

  @parameterized.named_parameters(
      dict(
          testcase_name='INT8_input',
          activation_tensor_config=_TensorQuantConfig(
              num_bits=8,
              symmetric=False,
              granularity=_Granularity.TENSORWISE,
          ),
          op=_OpName.INPUT,
      ),
      dict(
          testcase_name='INT8_output',
          activation_tensor_config=_TensorQuantConfig(
              num_bits=8,
              symmetric=False,
              granularity=_Granularity.TENSORWISE,
          ),
          op=_OpName.OUTPUT,
      ),
      dict(
          testcase_name='INT16_input',
          activation_tensor_config=_TensorQuantConfig(
              num_bits=16,
              symmetric=True,
              granularity=_Granularity.TENSORWISE,
          ),
          op=_OpName.INPUT,
      ),
      dict(
          testcase_name='INT16_output',
          activation_tensor_config=_TensorQuantConfig(
              num_bits=16,
              symmetric=True,
              granularity=_Granularity.TENSORWISE,
          ),
          op=_OpName.OUTPUT,
      ),
  )
  def test_input_output_explicit_set_quantize(
      self, activation_tensor_config, op
  ):
    self._quantizer.update_quantization_recipe(
        regex='.*',
        operation_name=op,
        op_config=_OpQuantConfig(
            activation_tensor_config=activation_tensor_config,
            weight_tensor_config=_TensorQuantConfig(
                num_bits=8,
                symmetric=True,
                granularity=_Granularity.TENSORWISE,
            ),
            compute_precision=_ComputePrecision.INTEGER,
        ),
    )
    calibration_result = {}
    if self._quantizer.need_calibration:
      calibration_result = self._quantizer.calibrate(_get_calibration_data())
    quantization_result = self._quantizer.quantize(calibration_result)
    quantized_model = tfl_flatbuffer_utils.read_model(
        bytearray(quantization_result.quantized_model)
    )
    self.assertLen(quantized_model.subgraphs, 1)
    subgraph = quantized_model.subgraphs[0]
    subgraph_tensors = subgraph.tensors
    self.assertLen(subgraph.inputs, 1)
    input_tensor = subgraph_tensors[subgraph.inputs[0]]
    output_tensor = subgraph_tensors[subgraph.outputs[0]]
    # Check input/output tensor type.
    if op == _OpName.INPUT:
      # See schema_py_generated.py for type code.
      if activation_tensor_config is None:
        self.assertEqual(input_tensor.type, 0)  # float32.
      elif activation_tensor_config.num_bits == 8:
        self.assertEqual(input_tensor.type, 9)  # int8.
      elif activation_tensor_config.num_bits == 16:
        self.assertEqual(input_tensor.type, 7)  # int16.
      self.assertEqual(output_tensor.type, 0)
    else:
      self.assertEqual(input_tensor.type, 0)  # float32.
      if activation_tensor_config is None:
        self.assertEqual(output_tensor.type, 0)  # float32.
      elif activation_tensor_config.num_bits == 8:
        self.assertEqual(output_tensor.type, 9)  # int8.
      elif activation_tensor_config.num_bits == 16:
        self.assertEqual(output_tensor.type, 7)  # int16.

    # Check accuracy.
    comparison_result = self._quantizer.validate(
        error_metrics='mse', test_data=_get_test_data()
    )
    self._check_comparison_result(comparison_result, output_tolerance=1e-4)

  @parameterized.parameters(
      ('../../recipes/default_a8w8_recipe.json', 9),
      ('../../recipes/default_a16w8_recipe.json', 7),
      ('../../recipes/default_af32w8float_recipe.json', 0),
      ('../../recipes/dynamic_wi8_afp32_recipe.json', 0),
      ('../../recipes/dynamic_legacy_wi8_afp32_recipe.json', 0),
  )
  def test_input_output_with_default_recipe(
      self, recipe_path, activation_type_code
  ):
    recipe_path = test_utils.get_path_to_datafile(recipe_path)
    self._quantizer.load_quantization_recipe(recipe_path)
    calibration_result = {}
    if self._quantizer.need_calibration:
      calibration_result = self._quantizer.calibrate(_get_calibration_data())
    quantization_result = self._quantizer.quantize(calibration_result)
    quantized_model = tfl_flatbuffer_utils.read_model(
        bytearray(quantization_result.quantized_model)
    )

    # Check input tensor type.
    self.assertLen(quantized_model.subgraphs, 1)
    subgraph = quantized_model.subgraphs[0]
    subgraph_tensors = subgraph.tensors
    self.assertLen(subgraph.inputs, 1)
    input_tensor = subgraph_tensors[subgraph.inputs[0]]
    output_tensor = subgraph_tensors[subgraph.outputs[0]]
    # See schema_py_generated.py for type code.
    self.assertEqual(input_tensor.type, activation_type_code)
    self.assertEqual(output_tensor.type, activation_type_code)
    # check accuracy.
    comparison_result = self._quantizer.validate(
        error_metrics='mse', test_data=_get_test_data()
    )
    self._check_comparison_result(comparison_result, output_tolerance=1e-4)

  def _check_comparison_result(self, comparison_result, output_tolerance):
    comparison_result = comparison_result.get_all_tensor_results()
    output_mse = comparison_result['PartitionedCall:0']
    self.assertLess(output_mse, output_tolerance)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Duplicate buffer transformation."""

from ai_edge_quantizer import qtyping
from ai_edge_quantizer.transformations import transformation_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils


def duplicate_buffer(
    transformation_input: transformation_utils.TransformationInput,
) -> qtyping.TransformationInfo:
  """Duplicates the buffer of the tensor."""
  tensor_id = transformation_input.tensor_id
  tensor = transformation_input.subgraph.tensors[tensor_id]
  buffer_data = transformation_input.buffers[tensor.buffer].data
  if buffer_data is None:
    tensor_name = tfl_flatbuffer_utils.get_tensor_name(tensor)
    raise ValueError(
        'Duplicate Buffer transformation supports only constant tensors.'
        f' Tensor {tensor_name} is not constant.'
    )

  duplicated_buffer_id = transformation_utils.get_constant_buffer(
      data=buffer_data,
      buffers=transformation_input.buffers,
      force_duplicate_buffer=True,
  )
  tensor.buffer = duplicated_buffer_id

  return qtyping.TransformationInfo(
      op_id=0, num_ops_added=0, output_tensor_id=tensor_id
  )



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import pathlib
import unittest  # pylint: disable=unused-import, required for OSS.

from absl.testing import parameterized
import absl.testing.absltest as absltest


from ai_edge_quantizer import quantizer
from ai_edge_quantizer import recipe
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils


_TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile('')


class RecipeTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    # Weights has < 1024 elements so legacy recipe will not quantize it.
    self._small_model_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH)
        / 'tests/models/single_conv2d_transpose_bias.tflite',
    )
    self._test_model_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH)
        / 'tests/models/conv_fc_mnist.tflite',
    )

  def _quantize_with_recipe_func(self, recipe_func, test_model_path):
    qt = quantizer.Quantizer(test_model_path)
    qt.load_quantization_recipe(recipe_func())
    self.assertIsNone(qt._result.quantized_model)
    if qt.need_calibration:
      calibration_data = tfl_interpreter_utils.create_random_normal_input_data(
          qt.float_model,
          num_samples=1,
      )
      calibration_result = qt.calibrate(calibration_data)
      quantization_result = qt.quantize(calibration_result)
    else:
      quantization_result = qt.quantize()
    self.assertIsNotNone(quantization_result.quantized_model)
    return quantization_result

  @unittest.skip('skipping due to b/438971945')
  def test_quantization_from_dynamic_wi8_afp32_func_succeeds(self):
    quant_result = self._quantize_with_recipe_func(
        recipe.dynamic_wi8_afp32, self._test_model_path
    )
    self.assertLess(
        len(quant_result.quantized_model),
        pathlib.Path(self._test_model_path).stat().st_size,
    )

  @unittest.skip('skipping due to b/438971945')
  def test_quantization_from_dynamic_wi4_afp32_func_succeeds(self):
    quant_result = self._quantize_with_recipe_func(
        recipe.dynamic_wi4_afp32, self._test_model_path
    )
    self.assertLess(
        len(quant_result.quantized_model),
        pathlib.Path(self._test_model_path).stat().st_size,
    )

  @unittest.skip('skipping due to b/438971945')
  def test_quantization_from_weight_only_wi8_afp32_func_succeeds(self):
    quant_result = self._quantize_with_recipe_func(
        recipe.weight_only_wi8_afp32, self._test_model_path
    )
    self.assertLess(
        len(quant_result.quantized_model),
        pathlib.Path(self._test_model_path).stat().st_size,
    )

  @unittest.skip('skipping due to b/438971945')
  def test_quantization_from_weight_only_wi4_afp32_func_succeeds(self):
    quant_result = self._quantize_with_recipe_func(
        recipe.weight_only_wi4_afp32, self._test_model_path
    )
    self.assertLess(
        len(quant_result.quantized_model),
        pathlib.Path(self._test_model_path).stat().st_size,
    )

  def test_quantization_from_dynamic_legacy_wi8_afp32_func_succeeds(self):
    quant_result = self._quantize_with_recipe_func(
        recipe.dynamic_legacy_wi8_afp32,
        self._small_model_path,
    )
    self.assertLen(
        quant_result.quantized_model,
        pathlib.Path(self._small_model_path).stat().st_size,
    )

  @parameterized.named_parameters(
      dict(
          testcase_name='dynamic_wi8_afp32',
          recipe_json_path='recipes/dynamic_wi8_afp32_recipe.json',
          recipe_func=recipe.dynamic_wi8_afp32,
      ),
      dict(
          testcase_name='weight_only_wi8_afp32',
          recipe_json_path='recipes/default_af32w8float_recipe.json',
          recipe_func=recipe.weight_only_wi8_afp32,
      ),
      dict(
          testcase_name='weight_only_wi4_afp32',
          recipe_json_path='recipes/default_af32w4float_recipe.json',
          recipe_func=recipe.weight_only_wi4_afp32,
      ),
      dict(
          testcase_name='dynamic_legacy_wi8_afp32',
          recipe_json_path='recipes/dynamic_legacy_wi8_afp32_recipe.json',
          recipe_func=recipe.dynamic_legacy_wi8_afp32,
      ),
      dict(
          testcase_name='a8w8',
          recipe_json_path='recipes/default_a8w8_recipe.json',
          recipe_func=recipe.static_wi8_ai8,
      ),
      dict(
          testcase_name='a16w8',
          recipe_json_path='recipes/default_a16w8_recipe.json',
          recipe_func=recipe.static_wi8_ai16,
      ),
  )
  @unittest.skip('skipping due to b/438971945')
  def test_recipe_func_and_json_matches(self, recipe_json_path, recipe_func):
    # Quantize with recipe from function in recipe module.
    quant_result_from_func = self._quantize_with_recipe_func(
        recipe_func, self._test_model_path
    )

    # Quantize with recipe from json file.
    qt_json = quantizer.Quantizer(self._test_model_path)
    json_recipe_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH) / recipe_json_path
    )
    qt_json.load_quantization_recipe(json_recipe_path)
    if qt_json.need_calibration:
      calibration_data = tfl_interpreter_utils.create_random_normal_input_data(
          qt_json.float_model,
          num_samples=1,
      )
      calibration_result = qt_json.calibrate(calibration_data)
      quant_result_from_json = qt_json.quantize(calibration_result)
    else:
      quant_result_from_json = qt_json.quantize()
    self.assertIsNotNone(quant_result_from_json.quantized_model)

    # Check if the quantized models match.
    self.assertEqual(
        len(quant_result_from_func.quantized_model),
        len(quant_result_from_json.quantized_model),
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Uniform quantize in tensor level."""

import dataclasses
from typing import Optional, Sequence
import ml_dtypes
import numpy as np
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.utils import tfl_flatbuffer_utils


@dataclasses.dataclass(frozen=True)
class IntType:
  num_bits: int
  signed: bool


def is_blockwise(granularity: qtyping.QuantGranularity) -> bool:
  """Checks if the quantization granularity is blockwise."""
  return "BLOCKWISE" in str(granularity)


def get_quantized_range(qtype: IntType) -> tuple[float, float]:
  """Calculates range of the quantized type."""
  if qtype.signed:
    qmax = 2 ** (qtype.num_bits - 1) - 1
    qmin = -(2 ** (qtype.num_bits - 1))
  else:
    qmax = (2**qtype.num_bits) - 1
    qmin = 0
  return float(qmin), float(qmax)


def extract_block_size_from_granularity(
    granularity: qtyping.QuantGranularity,
) -> int:
  """Get the block size for blockwise quantization."""
  if granularity == qtyping.QuantGranularity.BLOCKWISE_32:
    return 32
  elif granularity == qtyping.QuantGranularity.BLOCKWISE_64:
    return 64
  elif granularity == qtyping.QuantGranularity.BLOCKWISE_128:
    return 128
  elif granularity == qtyping.QuantGranularity.BLOCKWISE_256:
    return 256
  else:
    return 0


def _round_and_clip_inplace(
    tensor: np.ndarray, qtype: IntType, narrow: bool
) -> np.ndarray:
  """Round and clip the tensor to the given type in-place, but don't cast it."""
  qmin, qmax = get_quantized_range(qtype)
  if narrow:
    if qtype.signed:
      return np.clip(
          np.rint(tensor, out=tensor if not np.isscalar(tensor) else None),
          qmin + 1,
          qmax,
          out=tensor if not np.isscalar(tensor) else None,
      )
    else:
      raise ValueError("Unsigned data type should not have narrow range.")
  else:
    return np.clip(
        np.rint(tensor, out=tensor if not np.isscalar(tensor) else None),
        qmin,
        qmax,
        out=tensor if not np.isscalar(tensor) else None,
    )


def assign_quantized_type(tensor: np.ndarray, qtype: IntType) -> np.ndarray:
  """Cast the tensor to the quantized type."""
  if qtype.num_bits <= 8:
    qtype = np.int8 if qtype.signed else np.uint8
  elif qtype.num_bits <= 16:
    qtype = np.int16 if qtype.signed else np.uint16
  elif qtype.num_bits <= 32:
    qtype = np.int32 if qtype.signed else np.uint32
  else:
    qtype = np.int64 if qtype.signed else np.uint64
  return tensor.astype(qtype)


def fix_quantization_params_rank(
    tensor_data: np.ndarray,
    quantization_params: qtyping.UniformQuantParams,
) -> qtyping.UniformQuantParams:
  """Fix the rank of quantization parameters (scale/zero points).

  Scale and zero points need to be the same rank as tensor_data to avoid
  ambiguous broadcasting.

  Args:
    tensor_data: The tensor to be quantized.
    quantization_params: The quantization parameters.

  Returns:
    quantization_params with broadcasted scales and zero_points.
  """
  scales, zero_points = (
      quantization_params.scale,
      quantization_params.zero_point,
  )
  if tensor_data.ndim == scales.ndim:
    return quantization_params

  if tensor_data.ndim == 0:
    # Scalar tensor requires scalar scale and zero_point.
    if scales.size != 1 or zero_points.size != 1:
      raise ValueError(
          "Scale and zero_point must contain single element for scalar tensor."
          f" Got scale: {scales}, zero_point: {zero_points}"
      )
    scales = np.array(scales.item())
    zero_points = np.array(zero_points.item())
  else:
    dims = [
        dim
        for dim in range(tensor_data.ndim)
        if dim != quantization_params.quantized_dimension
    ]
    scales = np.expand_dims(scales, axis=dims)
    zero_points = np.expand_dims(zero_points, axis=dims)

  return qtyping.UniformQuantParams(
      scale=scales,
      zero_point=zero_points,
      num_bits=quantization_params.num_bits,
      symmetric=quantization_params.symmetric,
      quantized_dimension=quantization_params.quantized_dimension,
      quantized_data=quantization_params.quantized_data,
      block_size=quantization_params.block_size,
  )


def _get_tensor_shape_for_blockwise(
    tensor_shape: Sequence[int], quantized_dim: int, block_size: int
) -> list[int]:
  """Get the tensor shape for blockwise quantization.

  This function splits the quantize dimension of the tensor into blocks and the
  dim/blocks. Hence, min/max of the tensor can be calculated for each block
  using existing functions.

  Args:
    tensor_shape: The original shape of the tensor.
    quantized_dim: The dimension to be quantized blockwise.
    block_size: The size of the block.

  Returns:
    The new tensor shape for calculating scale and zp for blockwise
    quantization.
  """
  new_shape = []
  for index, val in enumerate(tensor_shape):
    if index == quantized_dim:
      if val % block_size != 0:
        raise ValueError(
            f"Quantized dimension {val} in tensor shape {tensor_shape} is not"
            f" divisible by block size {block_size}."
        )
      new_shape.append(int(val / block_size))
      new_shape.append(block_size)
    else:
      new_shape.append(val)
  return new_shape


def reshape_data_for_blockwise(
    tensor_data: np.ndarray,
    op_name: qtyping.TFLOperationName,
    granularity: qtyping.QuantGranularity,
) -> tuple[np.ndarray, int]:
  """Reshapes data for blockwise quantization.

  Args:
    tensor_data: The original tensor data.
    op_name: The name of the TFL op.
    granularity: The quantization granularity for the tensor.

  Returns:
    A tuple containing the reshaped tensor data and the new reduce dimension.
  """
  quantized_dim = tfl_flatbuffer_utils.TFL_OP_TO_BLOCKWISE_WEIGHT_QUANTIZED_DIM[
      op_name
  ]
  block_size = extract_block_size_from_granularity(granularity)
  new_shape = _get_tensor_shape_for_blockwise(
      tensor_data.shape, quantized_dim, block_size
  )
  return tensor_data.reshape(new_shape), quantized_dim + 1


def _broadcast_scale_zp_for_blockwise(
    tensor_content: np.ndarray,
    quant_params: qtyping.UniformQuantParams,
) -> qtyping.UniformQuantParams:
  """Broadcasts scale and zp for blockwise quantization.

  Args:
    tensor_content: The original tensor data.
    quant_params: The quantization parameters.
      `quant_params.quantized_dimension` must be specified.
      `quant_params.block_size` must be specified and positive.

  Returns:
    The updated quantization parameters with broadcasted scale and zp for
    correct constant quantization.
  """
  if quant_params.quantized_dimension is None:
    raise ValueError("Quantized dimension must be specified.")
  if quant_params.block_size is None or quant_params.block_size <= 0:
    raise ValueError("Block size must be specified and positive.")
  quantized_dim = quant_params.quantized_dimension
  expanded_tensor_shape = _get_tensor_shape_for_blockwise(
      tensor_content.shape, quantized_dim, quant_params.block_size
  )
  expanded_scale = np.reshape(
      np.broadcast_to(
          np.expand_dims(quant_params.scale, quantized_dim + 1),
          expanded_tensor_shape,
      ),
      tensor_content.shape,
  )
  if quant_params.zero_point is None or quant_params.zero_point.size == 0:
    expanded_zp = np.zeros_like(tensor_content, dtype=np.int32)
  else:
    expanded_zp = np.reshape(
        np.broadcast_to(
            np.expand_dims(quant_params.zero_point, quantized_dim + 1),
            expanded_tensor_shape,
        ),
        tensor_content.shape,
    )
  return qtyping.UniformQuantParams(
      scale=expanded_scale,
      zero_point=expanded_zp,
      num_bits=quant_params.num_bits,
      symmetric=quant_params.symmetric,
      quantized_dimension=quantized_dim,
      block_size=quant_params.block_size,
  )


def uniform_quantize(
    tensor_data: np.ndarray,
    quantization_params: qtyping.UniformQuantParams,
    is_blockwise_quant: bool = False,
):
  """Uniform quantize a tensor.

  Args:
    tensor_data: The tensor to be quantized.
    quantization_params: The quantization parameters.
    is_blockwise_quant: Whether the tensor is blockwise quantized.

  Returns:
    The quantized tensor.
  """
  # The reshaping for blockwise quantization is unique hence we do this here
  # to avoid unexpected broadcast behavior downstream.
  if is_blockwise_quant:
    quantization_params = _broadcast_scale_zp_for_blockwise(
        tensor_data, quantization_params
    )

  # quant params in flatbuffer is flattened, expand the rank to be the same
  # as the tensor rank to avoid ambiguous broadcasting.
  quantization_params = fix_quantization_params_rank(
      tensor_data, quantization_params
  )
  _is_valid_quantization_params(tensor_data, quantization_params)
  scales, zero_points = (
      quantization_params.scale,
      quantization_params.zero_point,
  )
  # TODO: b/332574603 - support unsigned data type.
  qtype = IntType(quantization_params.num_bits, signed=True)
  # For quantization with more than 8 bits, symmetric narrow-range quantization
  # is required due to assumptions made by legacy TFLite kernels. However, this
  # method is not ideal for low-bit quantization (e.g., 2-bit quantization,
  # which only has 4 bins), as it wastes a bin and there are no kernel
  # requirements for a narrow range when < 8 bits because the data is unpacked
  # to int8 before being used in the kernel.
  narrow_range = (
      quantization_params.symmetric and quantization_params.num_bits >= 8
  )
  required_dtype = np.signedinteger if qtype.signed else np.unsignedinteger
  if not np.issubdtype(zero_points.dtype, required_dtype):
    raise ValueError(
        f"zero_points need to be {required_dtype}."
        f" But the actual type is {zero_points.dtype}."
    )
  ret = np.divide(tensor_data, scales)
  ret = np.add(ret, zero_points, out=ret if not np.isscalar(ret) else None)
  ret = _round_and_clip_inplace(ret, qtype, narrow_range)
  ret = assign_quantized_type(ret, qtype)
  return ret


def uniform_dequantize(
    tensor_data: np.ndarray,
    quantization_params: qtyping.UniformQuantParams,
):
  """Uniform dequantize a tensor.

  Args:
    tensor_data: The tensor to be dequantized.
    quantization_params: The quantization parameters.

  Returns:
    The dequantized tensor.
  """
  if quantization_params.block_size != 0:
    # b/443830202: The quantized dimension is currently increased by 1 because
    # AEQ expects 1 and XNNPack expects 0.
    quantization_params = dataclasses.replace(
        quantization_params,
        quantized_dimension=quantization_params.quantized_dimension + 1,
    )
    scale_shape = list(tensor_data.shape)
    scale_shape[quantization_params.quantized_dimension] = (
        scale_shape[quantization_params.quantized_dimension]
        // quantization_params.block_size
    )
    quantization_params = dataclasses.replace(
        quantization_params,
        scale=quantization_params.scale.reshape(scale_shape),
    )
    quantization_params = _broadcast_scale_zp_for_blockwise(
        tensor_data, quantization_params
    )

  # quant params in flatbuffer is flattened, expand the rank to be the same
  # as the tensor rank to avoid ambiguous broadcasting.
  quantization_params = fix_quantization_params_rank(
      tensor_data, quantization_params
  )
  _is_valid_quantization_params(tensor_data, quantization_params)
  return np.multiply(
      tensor_data - quantization_params.zero_point, quantization_params.scale
  )


def symmetric_quantize_bias_tensor(
    bias_content: np.ndarray,
    input_tensor_quant_params: qtyping.UniformQuantParams,
    weight_tensor_quant_params: qtyping.UniformQuantParams,
    # TODO: b/477623376 - Have a debugging mode to check bias.
    check_error: bool = False,
) -> qtyping.UniformQuantParams:
  """Quantize bias tensor (symmetrically, i.e., zero_point = 0).

  We quantize bias to a much higher bit width, e.g., int32 for int8 weights. We
  can afford the cost of being symmetric all the time. This configuration fits
  TFL kernel designs.

  Args:
    bias_content: The bias content.
    input_tensor_quant_params: The quantization parameters of input tensor.
    weight_tensor_quant_params: The quantization parameters of weight tensor.
    check_error: Whether to check if the quantization error (the difference
      between the original and dequantized bias) is larger than the quantization
      scale. This check is important because bias quantization parameters are
      fixed (bias_scale = input_scale * weight_scale), which can lead to large
      quantization errors. Raising an error when the quantization error is
      larger than the scale helps to identify unexpected numerical issues.

  Returns:
    The quantized bias tensor.
  """
  input_tensor_scale = input_tensor_quant_params.scale
  weight_tensor_scale = weight_tensor_quant_params.scale
  # Bias is always 1D, make sure the scale has 1D shape as well.
  effective_output_scale = np.squeeze(input_tensor_scale * weight_tensor_scale)
  # Squeeze can produce scalar, but we want 1D tensor.
  if not effective_output_scale.shape:
    effective_output_scale = np.expand_dims(effective_output_scale, axis=0)

  # symmetric
  bias_zp = np.zeros_like(effective_output_scale, dtype=np.int32)
  # Fixed to 32 bits since most of the accelerators use int32 accumulator.
  bias_number_bits = 32
  symmetric = True
  quantized_dimension = None if len(effective_output_scale) == 1 else 0
  bias_quant_params = qtyping.UniformQuantParams(
      scale=effective_output_scale,
      zero_point=bias_zp,
      num_bits=bias_number_bits,
      symmetric=symmetric,
      quantized_dimension=quantized_dimension,
  )

  quantized_vars = uniform_quantize(bias_content, bias_quant_params)
  if check_error:
    dequantized_bias = uniform_dequantize(quantized_vars, bias_quant_params)
    max_quant_error = np.max(np.abs(dequantized_bias - bias_content))
    error_tolerance = np.maximum(1e-6, np.max(effective_output_scale))
    if max_quant_error > error_tolerance:
      raise ValueError(
          "Quantization error is too large for bias tensor quantization. Max"
          f" quantization error is {max_quant_error}, which exceed"
          f" the threshold {error_tolerance}"
      )

  # Save the int32 quantized bias as int64 if the input tensor is quantized to
  # 16 bits. This is to assume the matmul is using int64 accumulator (safe from
  # overflow). For accelerators with int32 accumulator, it is safe to cast int64
  # back to int32.
  if input_tensor_quant_params.num_bits == 16:
    quantized_vars = quantized_vars.astype(np.int64)
    bias_number_bits = 64

  # UniformQuantParams is frozen dataclass, need to recreate.
  return qtyping.UniformQuantParams(
      scale=effective_output_scale,
      zero_point=bias_zp,
      num_bits=bias_number_bits,
      quantized_dimension=quantized_dimension,
      symmetric=symmetric,
      quantized_data=quantized_vars,
  )


def tensor_zp_scale_from_min_max(
    min_value,
    max_value,
    num_bits: int,
    symmetric: bool,
    granularity: qtyping.QuantGranularity,
    clipping_values: Optional[np.ndarray] = None,
):
  """Get zero point and scale from min and max value.

  Args:
    min_value: The minimum value of the tensor (channelwise and blockwise
      supported).
    max_value: The maximum value of the tensor (channelwise and blockwise
      supported).
    num_bits: The number of bits of the tensor.
    symmetric: Whether the tensor is symmetric.
    granularity: The granularity of the tensor.
    clipping_values: Absolute clipping values to apply to the tensor. This will
      clip the tensors to the range [-clipping_values, clipping_values]. This
      should be the same shape as min_value and max_value. If None, no clipping
      will be applied.

  Returns:
    The zero point and scale of the tensor.
  """

  # TODO: b/332574603 - support unsigned data type.
  qtype = IntType(
      num_bits,
      signed=True,
  )
  qmin, qmax = get_quantized_range(qtype)
  min_bound = 1e-9  # Avoid zero scale.
  pos_clipping_values = None if clipping_values is None else clipping_values
  neg_clipping_values = None if clipping_values is None else -clipping_values

  if is_blockwise(granularity):
    # Blockwise quantization uses float16 scale,
    # with 7 bit mantissa, so the maximum scale value is 65280 and maximum
    # representable range is [-65280 * (2 ** num_bits),
    # 65280 * (2 ** num_bits - 1)].
    # Note that we have one extra value on the negative side.
    float16_max = np.broadcast_to(
        np.array(65280) * (2**num_bits - 1), max_value.shape
    )
    float16_min = np.broadcast_to(
        np.array(-65280) * (2**num_bits), min_value.shape
    )
    pos_clipping_values = (
        float16_max
        if pos_clipping_values is None
        else np.minimum(pos_clipping_values, float16_max)
    )
    neg_clipping_values = (
        float16_min
        if neg_clipping_values is None
        else np.maximum(neg_clipping_values, float16_min)
    )

  if symmetric:
    bound = np.maximum(np.abs(min_value), np.abs(max_value))
    bound = np.maximum(bound, min_bound)
    if clipping_values is not None:
      bound = np.clip(bound, neg_clipping_values, pos_clipping_values)
    if not qtype.signed:
      half_q = (qmax - 1) / 2
      scale = bound / half_q
      zp = np.ones_like(scale) * (half_q + 1)
    else:
      scale = bound / qmax
      zp = np.zeros_like(scale, dtype=np.int32)
  else:
    # Include 0 to the range to support zero-padding.
    # See: https://arxiv.org/pdf/1712.05877.pdf
    # This ensures bound_min <= 0 <= bound_max.
    bound_max = np.maximum(max_value, np.zeros_like(max_value))
    bound_min = np.minimum(min_value, np.zeros_like(min_value))
    bound = np.maximum(bound_max - bound_min, min_bound)
    if clipping_values is not None:
      bound = np.clip(bound, -clipping_values, clipping_values)
    scale = bound / (qmax - qmin)
    zp = qmin - bound_min / scale
    zp = np.rint(zp)

  if is_blockwise(granularity):
    # Round the scale values to 7 bit mantissa.
    scale = (
        scale.astype(ml_dtypes.bfloat16).astype(np.float16).astype(np.float32)
    )

  # It's safe to cast zp to qtype without clipping because we can infer
  # qmin <= zp <= qmax from bound_min <= 0 <= bound_max.
  zp = assign_quantized_type(zp, qtype)
  return zp, scale


def _is_valid_quantization_params(
    tensor_data: np.ndarray,
    quantization_params: qtyping.UniformQuantParams,
) -> None:
  """Checks if the quantization parameters are valid.

  A valid quantization params requires:
    1. scale and zero point either have the same shape or the zero point is a
    scalar.
    2. scale and zero point have the same rank as the tensor content (avoid
    ambiguous broadcasting).

  Args:
    tensor_data: The tensor to be quantized.
    quantization_params: The quantization parameters.

  Returns:
    True if the quantization parameters are valid.
  """
  if (
      quantization_params.scale.shape != quantization_params.zero_point.shape
      and quantization_params.zero_point.size != 1
  ):
    raise ValueError(
        "scale and zero_point must have the same shape or zero_point must have"
        f" only one element. Got {quantization_params.scale.shape} and"
        f" {quantization_params.zero_point.shape}"
    )

  tensor_rank = tensor_data.ndim
  scale_rank = quantization_params.scale.ndim
  zero_point_rank = quantization_params.zero_point.ndim
  if tensor_rank != scale_rank or (tensor_rank != zero_point_rank):
    raise ValueError(
        f"Ranks of scales ({scale_rank}) and zps"
        f" ({zero_point_rank}) must be the same as the tensor rank"
        f" ({tensor_rank})."
    )
  if (
      quantization_params.block_size != 0
      and tensor_data.shape[quantization_params.quantized_dimension]
      % quantization_params.block_size
      != 0
  ):
    raise ValueError(
        "Tensor dimension must be divisible by block size. Got dimension:"
        f" {tensor_data.shape[quantization_params.quantized_dimension]} and"
        f" block size: {quantization_params.block_size}"
    )



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Implements the Hadamard Rotation quantization."""

from typing import Any, Optional
import numpy as np
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import octav
from ai_edge_quantizer.algorithms.utils import common_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils


CUSTOM_OP_ALGORITHM_KEY = "HADAMARD_ROTATION"
DECOMPOSED_ALGORITHM_KEY = "DECOMPOSED_HADAMARD_ROTATION"


def _make_hadamard_matrix(size: int) -> np.ndarray:
  """Generates a Hadamard matrix of the given size.

  Args:
    size: The size of the Hadamard matrix. Must be a power of 2. This represents
      a single dimension. E.g. if size is 4, then the Hadamard matrix is a 4x4
      matrix.

  Returns:
    The Hadamard matrix.

  Raises:
    ValueError: If the size is not a power of 2.
  """
  if size <= 0 or (size & (size - 1)) != 0:
    raise ValueError("Hadamard matrix size must be a power of 2. ")
  h = h2 = np.array([[1, 1], [1, -1]])
  current_size = 2
  while current_size < size:
    h = np.kron(h, h2)
    current_size *= 2
  return h / np.sqrt(size)


def _rotate_with_diagonal_hadamard(
    tensor_content: np.ndarray,
    axis: int,
):
  """Quantizes the given float array using the diagonal Hadamard algorithm.

  Args:
    tensor_content: The float array to quantize.
    axis: The axis of the tensor to rotate.

  Returns:
    A tuple containing the quantized array and the recovered array.

  Raises:
    ValueError: If the axis is not the last axis of tensor_content. To support
      other axes, please add support to the matrix multiplication.
  """
  if axis != tensor_content.ndim - 1:
    raise ValueError(
        "Hadamard rotation is only supported for tensors with quantized"
        " dimension 0 (rotate last dimension)."
    )

  # Use the largest power of 2 that is a factor of the dimension and then
  # tile this Hadamard matrix along the diagonal. 2**30 is just a large power
  # of 2 to calculate this factor.
  hadamard_size = np.gcd(tensor_content.shape[axis], 2 ** 30)
  diagonal_size = tensor_content.shape[axis] // hadamard_size
  # Output size is the product of all dimensions except the one being rotated.
  output_size = np.prod(np.delete(tensor_content.shape, axis))
  random_vector = np.ones(hadamard_size, dtype=np.int8)

  # Use a canonical Hadamard matrix.
  hadamard = _make_hadamard_matrix(hadamard_size)
  reshaped_tensor = tensor_content.reshape(
      diagonal_size * output_size, hadamard_size)
  w_rotated = np.matmul(reshaped_tensor, hadamard)
  return w_rotated.reshape(tensor_content.shape), hadamard_size, random_vector


def get_tensor_quant_params(
    op_info: qtyping.OpInfo,
    tensor_quant_config: qtyping.TensorQuantizationConfig,
    tensor_content: Optional[np.ndarray] = None,
    tensor_qsv: Optional[dict[str, Any]] = None,
) -> qtyping.UniformQuantParams:
  """Returns the quantization parameters for a tensor.

  This function will rotate the tensor with a Hadamard matrix and then
  quantize it with OCTAV.

  Args:
    op_info: Aggregated information about the op (e.g., quantization config).
    tensor_quant_config: The quantization config for the tensor.
    tensor_content: The content of the tensor. When None, it means the tensor is
      not a weight tensor (e.g. static quantization).
    tensor_qsv: A dictionary containing the min/max of the tensor.

  Raises:
    ValueError: If the blockwise quantization is requested.
    ValueError: If the asymmetric quantization is requested.
    ValueError: `tensor_qsv` must contain min/max values, or `tensor_content`
      must be provided so that they can be inferred.
  """
  if tensor_content is None:
    raise ValueError("Hadamard rotation is only supported for weight tensors.")

  if tensor_qsv is not None:
    raise ValueError(
        "Hadamard rotation is not supported for static quantization."
    )

  if tensor_content.ndim < 2:
    raise ValueError(
        "Hadamard rotation is only supported for tensors with rank >= 2."
    )

  # Reduction axis is the last non-quantized dimension. Since we only support
  # quantized_dim of 0 (or 1 for blockwise), the reduction axis is the last
  # axis.
  reduce_axis = tensor_content.ndim - 1

  # Rotate the tensor with a Hadamard matrix.
  w_rotated, hadamard_size, random_vector = _rotate_with_diagonal_hadamard(
      tensor_content, axis=reduce_axis
  )

  # Get the quantized values of the rotated tensor.
  qparams = octav.get_tensor_quant_params(
      op_info, tensor_quant_config, w_rotated, tensor_qsv
  )

  return qtyping.UniformQuantParams(
      quantized_dimension=qparams.quantized_dimension,
      num_bits=qparams.num_bits,
      scale=qparams.scale,
      zero_point=qparams.zero_point,
      symmetric=qparams.symmetric,
      quantized_data=qparams.quantized_data,
      block_size=qparams.block_size,
      hadamard=qtyping.UniformQuantParams.HadamardRotationParams(
          random_binary_vector=random_vector,
          hadamard_size=hadamard_size,
      ),
  )


def _materialize_fully_connected(
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_quant_params_cache: common_utils.TensorQuantParamsCache,
    is_decomposed: bool = False,
    tensor_name_to_qsv: Optional[dict[str, Any]] = None,  # pylint: disable=unused-argument
) -> list[qtyping.TensorTransformationParams]:
  """Materialize the fully_connected op.

  Args:
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    tensor_quant_params_cache: Cache of already computed
      `UniformQuantParams|NonLinearQuantParams` objects keyed on a tuple of the
      buffer ID and the `TensorQuantizationConfig` used to compute it.
    is_decomposed: Whether to use decomposed Hadamard rotation ops or a custom
      op.
    tensor_name_to_qsv: A map of tensor name to quantization parameters.

  Returns:
    Quantization configuration for the tensors associated with the op (e.g.,
    weights, bias).
  """
  if op_info.op_quant_config.weight_tensor_config is None:
    raise ValueError(
        "Weight tensor quantization config is not provided for Hadamard"
        " Rotation quantization."
    )

  op_tensor_params = []

  # Materialize weight.
  weight_tensor_index = 1
  weight_tensor = graph_info.subgraph_tensors[
      op_info.op.inputs[weight_tensor_index]
  ]
  tensor_data = tfl_flatbuffer_utils.get_tensor_data(
      weight_tensor, graph_info.buffers
  )

  if not (
      quant_params := tensor_quant_params_cache.lookup(
          weight_tensor.buffer, op_info.op_quant_config.weight_tensor_config
      )
  ):
    # quant_params contains the rotated and quantized weights done by
    # get_tensor_quant_params().
    quant_params = get_tensor_quant_params(
        op_info,
        op_info.op_quant_config.weight_tensor_config,
        tensor_data,
        None,
    )
    tensor_quant_params_cache.insert(
        weight_tensor.buffer,
        op_info.op_quant_config.weight_tensor_config,
        quant_params,
    )

  transformations = [qtyping.QuantTransformation.QUANTIZE_TENSOR]
  op2tensor_params = qtyping.OpToTensorParams(
      subgraph_op_id=op_info.subgraph_op_index,
      parameters=quant_params,
      transformations=transformations,
  )
  weight_transformation_params = qtyping.TensorTransformationParams(
      tensor_name=tfl_flatbuffer_utils.get_tensor_name(weight_tensor),
      consumers=[op2tensor_params],
  )

  # Materialize input. A hadamard rotation op should be inserted on the input
  # tensor to do the inverse of the weight's transformation.
  input_tensor_index = 0
  input_tensor = graph_info.subgraph_tensors[
      op_info.op.inputs[input_tensor_index]
  ]
  transformations = [
      qtyping.QuantTransformation.INSERT_DECOMPOSED_HADAMARD_ROTATION
      if is_decomposed
      else qtyping.QuantTransformation.INSERT_HADAMARD_ROTATION,
  ]
  op2tensor_params = qtyping.OpToTensorParams(
      subgraph_op_id=op_info.subgraph_op_index,
      parameters=quant_params,
      transformations=transformations,
  )
  input_transformation_params = qtyping.TensorTransformationParams(
      tensor_name=tfl_flatbuffer_utils.get_tensor_name(input_tensor),
      consumers=[op2tensor_params],
  )
  op_tensor_params.append(input_transformation_params)
  op_tensor_params.append(weight_transformation_params)

  # Materialize bias. Since static quantization is not supported, we do not
  # quantize the bias tensor.
  bias_tensor_index = 2
  bias_tensor = graph_info.subgraph_tensors[
      op_info.op.inputs[bias_tensor_index]
  ]
  no_quant_tensor_params = qtyping.OpToTensorParams(
      subgraph_op_id=op_info.subgraph_op_index,
      transformations=[qtyping.QuantTransformation.NO_QUANTIZE],
  )
  bias_transformation_params = qtyping.TensorTransformationParams(
      tensor_name=tfl_flatbuffer_utils.get_tensor_name(bias_tensor),
      consumers=[no_quant_tensor_params],
  )
  op_tensor_params.append(bias_transformation_params)

  # Materialize output. Since static quantization is not supported, we do not
  # quantize the output tensor.
  output_tensor_index = 0
  output_tensor = graph_info.subgraph_tensors[
      op_info.op.outputs[output_tensor_index]
  ]
  no_quant_tensor_params = qtyping.OpToTensorParams(
      subgraph_op_id=op_info.subgraph_op_index,
      transformations=[qtyping.QuantTransformation.NO_QUANTIZE],
  )
  output_transformation_params = qtyping.TensorTransformationParams(
      tensor_name=tfl_flatbuffer_utils.get_tensor_name(output_tensor),
      producer=no_quant_tensor_params,
  )
  op_tensor_params.append(output_transformation_params)

  return op_tensor_params


def materialize_fully_connected_custom_op(
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_quant_params_cache: common_utils.TensorQuantParamsCache,
    tensor_name_to_qsv: Optional[dict[str, Any]] = None,
) -> list[qtyping.TensorTransformationParams]:
  return _materialize_fully_connected(
      op_info,
      graph_info,
      is_decomposed=False,
      tensor_name_to_qsv=tensor_name_to_qsv,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_fully_connected_decomposed(
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_quant_params_cache: common_utils.TensorQuantParamsCache,
    tensor_name_to_qsv: Optional[dict[str, Any]] = None,
) -> list[qtyping.TensorTransformationParams]:
  return _materialize_fully_connected(
      op_info,
      graph_info,
      is_decomposed=True,
      tensor_name_to_qsv=tensor_name_to_qsv,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def _materialize_embedding_lookup(
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_quant_params_cache: common_utils.TensorQuantParamsCache,
    is_decomposed: bool = False,
    tensor_name_to_qsv: Optional[dict[str, Any]] = None,  # pylint: disable=unused-argument
) -> list[qtyping.TensorTransformationParams]:
  """Materialize the embedding_lookup op.

  Args:
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    tensor_quant_params_cache: Cache of already computed
      `UniformQuantParams|NonLinearQuantParams` objects keyed on a tuple of the
      buffer ID and the `TensorQuantizationConfig` used to compute it.
    is_decomposed: Whether to use decomposed Hadamard rotation ops or a custom
      op.
    tensor_name_to_qsv: A map of tensor name to quantization parameters.

  Returns:
    Quantization configuration for the tensors associated with the op (e.g.,
    weights, bias).
  """
  op_tensor_params = []

  # Materialize lookup.
  lookup_tensor_index = 0
  lookup_tensor = graph_info.subgraph_tensors[
      op_info.op.inputs[lookup_tensor_index]
  ]
  transformations = [
      qtyping.QuantTransformation.NO_QUANTIZE,
  ]
  op2tensor_params = qtyping.OpToTensorParams(
      subgraph_op_id=op_info.subgraph_op_index,
      parameters=None,
      transformations=transformations,
  )
  lookup_transformation_params = qtyping.TensorTransformationParams(
      tensor_name=tfl_flatbuffer_utils.get_tensor_name(lookup_tensor),
      consumers=[op2tensor_params],
  )
  op_tensor_params.append(lookup_transformation_params)

  # Materialize embedding. The embedding table should be rotated and then
  # quantized.
  embedding_tensor_index = 1
  embedding_tensor = graph_info.subgraph_tensors[
      op_info.op.inputs[embedding_tensor_index]
  ]
  tensor_data = tfl_flatbuffer_utils.get_tensor_data(
      embedding_tensor, graph_info.buffers
  )

  if tensor_quant_params_cache is None or not (
      quant_params := tensor_quant_params_cache.lookup(
          embedding_tensor.buffer, op_info.op_quant_config.weight_tensor_config
      )
  ):
    quant_params = get_tensor_quant_params(
        op_info,
        op_info.op_quant_config.weight_tensor_config,
        tensor_data,
        None,
    )
    if tensor_quant_params_cache:
      tensor_quant_params_cache.insert(
          embedding_tensor.buffer,
          op_info.op_quant_config.weight_tensor_config,
          quant_params,
      )

  transformations = [qtyping.QuantTransformation.QUANTIZE_TENSOR]
  op2tensor_params = qtyping.OpToTensorParams(
      subgraph_op_id=op_info.subgraph_op_index,
      parameters=quant_params,
      transformations=transformations,
  )
  weight_transformation_params = qtyping.TensorTransformationParams(
      tensor_name=tfl_flatbuffer_utils.get_tensor_name(embedding_tensor),
      consumers=[op2tensor_params],
  )
  op_tensor_params.append(weight_transformation_params)

  # Materialize output. A hadamard rotation op should be inserted on the output
  # tensor to do the inverse of the embedding's transformation.
  output_tensor_index = 0
  output_tensor = graph_info.subgraph_tensors[
      op_info.op.outputs[output_tensor_index]
  ]
  transformations = [
      qtyping.QuantTransformation.INSERT_DECOMPOSED_HADAMARD_ROTATION
      if is_decomposed
      else qtyping.QuantTransformation.INSERT_HADAMARD_ROTATION,
  ]
  op2tensor_params = qtyping.OpToTensorParams(
      subgraph_op_id=op_info.subgraph_op_index,
      parameters=quant_params,
      transformations=transformations,
  )
  output_transformation_params = qtyping.TensorTransformationParams(
      tensor_name=tfl_flatbuffer_utils.get_tensor_name(output_tensor),
      producer=op2tensor_params,
  )
  op_tensor_params.append(output_transformation_params)

  return op_tensor_params


def materialize_embedding_lookup_custom_op(
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_quant_params_cache: common_utils.TensorQuantParamsCache,
    tensor_name_to_qsv: Optional[dict[str, Any]] = None,
) -> list[qtyping.TensorTransformationParams]:
  return _materialize_embedding_lookup(
      op_info,
      graph_info,
      is_decomposed=False,
      tensor_name_to_qsv=tensor_name_to_qsv,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_embedding_lookup_decomposed(
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_quant_params_cache: common_utils.TensorQuantParamsCache,
    tensor_name_to_qsv: Optional[dict[str, Any]] = None,
) -> list[qtyping.TensorTransformationParams]:
  return _materialize_embedding_lookup(
      op_info,
      graph_info,
      is_decomposed=True,
      tensor_name_to_qsv=tensor_name_to_qsv,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utility functions to display a progress bar and progress report."""

import time
import tracemalloc
import tqdm


class ProgressBar:
  """A Progress Bar that can be used to track the progress of a process."""

  def __init__(
      self,
      total_steps: int,
      description: str = '',
      disappear_on_finish: bool = False,
      disable: bool = False,
  ):
    self._progress_bar = tqdm.tqdm(
        total=total_steps,
        desc=description,
        leave=not disappear_on_finish,
        disable=disable,
    )

  def __enter__(self):
    """Enters the context manager."""
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    """Exits the context manager and closes the progress bar."""
    self.close()

  def update_single_step(self):
    """Updates the progress bar by a single step."""
    self._progress_bar.update(1)

  def close(self):
    """Closes the progress bar."""
    self._progress_bar.close()


class ProgressReport:
  """A class to generate a progress report for the quantization process."""

  def __init__(self, description: str = ''):
    self._description = description
    self._start_time = None

  def capture_progess_start(self):
    self._start_time = time.time()
    tracemalloc.start()

  def render_report(
      self,
      original_size: int,
      quantized_size: int,
      quantization_ratio: float,
      memory_peak: float,
      total_time: float,
  ):
    """Prints out the progress report."""
    print(f'Original model size: {original_size/1024:.2f} KB')
    print(f'Quantized model size: {quantized_size/1024:.2f} KB')
    print(f'Quantization Ratio: {quantization_ratio:.2f}')
    print(f'Total time: {total_time:.2f} seconds')
    print(f'Memory peak: {memory_peak:.2f} MB')

  def generate_progress_report(self, original_model, quantized_model):
    original_size = len(original_model)
    quantized_size = len(quantized_model)
    quantization_ratio = quantized_size / original_size
    total_time = time.time() - self._start_time
    _, mem_peak_bytes = tracemalloc.get_traced_memory()
    mem_peak_mb = mem_peak_bytes / 1024 / 1024
    self.render_report(
        original_size,
        quantized_size,
        quantization_ratio,
        mem_peak_mb,
        total_time,
    )



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Tests for transformation_utils."""

import pathlib
import numpy as np
from absl.testing import parameterized
import absl.testing.absltest as absltest
from ai_edge_quantizer.transformations import transformation_utils
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils
from ai_edge_litert import schema_py_generated  # pylint: disable=g-direct-tensorflow-import

TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile("../tests/models")


class TransformationUtilsTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH) / "single_fc_bias.tflite"
    )
    self.model = tfl_flatbuffer_utils.read_model(self.model_path)

  @parameterized.named_parameters(
      dict(
          testcase_name="add_new_op_code",
          op_code=schema_py_generated.BuiltinOperator.LOGISTIC,
          expected=1,
          custom_op_name=None,
      ),
      dict(
          testcase_name="add_existing_op_code",
          op_code=schema_py_generated.BuiltinOperator.FULLY_CONNECTED,
          expected=0,
          custom_op_name=None,
      ),
      dict(
          testcase_name="add_new_custom_op_code",
          op_code=schema_py_generated.BuiltinOperator.CUSTOM,
          expected=1,
          custom_op_name="random_new_custom_op",
      ),
  )
  def test_add_op_code(self, op_code, expected, custom_op_name):
    """Tests if the op code is added to the model."""
    got = transformation_utils.add_op_code(
        op_code=op_code,
        model_op_codes=self.model.operatorCodes,
        custom_op_name=custom_op_name,
    )
    self.assertEqual(expected, got)
    if custom_op_name is not None:
      self.assertEqual(self.model.operatorCodes[got].customCode, custom_op_name)

  def test_add_custom_op_code_without_op_string_raises_error(self):
    with self.assertRaisesRegex(ValueError, "Custom string is required"):
      transformation_utils.add_op_code(
          op_code=schema_py_generated.BuiltinOperator.CUSTOM,
          model_op_codes=self.model.operatorCodes,
          custom_op_name=None,
      )

  def test_add_two_custom_op_codes(self):
    custom_op_name = "random_new_custom_op"
    added_index = transformation_utils.add_op_code(
        op_code=schema_py_generated.BuiltinOperator.CUSTOM,
        model_op_codes=self.model.operatorCodes,
        custom_op_name=custom_op_name,
    )
    self.assertEqual(1, added_index)
    self.assertEqual(
        self.model.operatorCodes[added_index].customCode, custom_op_name
    )

    custom_op_name_2 = "random_new_custom_op_2"
    added_index = transformation_utils.add_op_code(
        op_code=schema_py_generated.BuiltinOperator.CUSTOM,
        model_op_codes=self.model.operatorCodes,
        custom_op_name=custom_op_name_2,
    )
    self.assertEqual(2, added_index)
    self.assertEqual(
        self.model.operatorCodes[added_index].customCode, custom_op_name_2
    )

  @parameterized.named_parameters(
      dict(
          testcase_name="float32",
          data=np.array([1.0, 2.0, 3.0, 4.0], dtype=np.float32),
      ),
      dict(
          testcase_name="int8",
          data=np.array([[1, 2], [3, 4]], dtype=np.int8),
      ),
  )
  def test_add_new_constant_buffer(self, data):
    """Tests if the constant buffer is added to the model."""
    prev_num_buffers = len(self.model.buffers) - 1
    new_buffer_idx = transformation_utils.get_constant_buffer(
        data=data,
        buffers=self.model.buffers,
    )
    self.assertEqual(new_buffer_idx, prev_num_buffers + 1)

    expected_buffer_data = (
        np.frombuffer(
            data.tobytes(),
            dtype=np.uint8,
        )
        .flatten()
        .tolist()
    )
    self.assertEqual(
        self.model.buffers[new_buffer_idx].data.tolist(),
        expected_buffer_data,
    )

  @parameterized.named_parameters(
      dict(
          testcase_name="float32",
          tensor_data=np.array([1.0, 2.0, 3.0, 4.0], dtype=np.float32),
          tensor_type=schema_py_generated.TensorType.FLOAT32,
          expected_type=schema_py_generated.TensorType.FLOAT32,
          expected_shape=(4,),
          expected_buffer_data=np.frombuffer(
              np.array([1.0, 2.0, 3.0, 4.0], dtype=np.float32).tobytes(),
              dtype=np.uint8,
          ).flatten(),
      ),
      dict(
          testcase_name="int8",
          tensor_data=np.array([[1, 2], [3, 4]], dtype=np.int8),
          tensor_type=schema_py_generated.TensorType.INT8,
          expected_type=schema_py_generated.TensorType.INT8,
          expected_shape=(2, 2),
          expected_buffer_data=np.frombuffer(
              np.array([[1, 2], [3, 4]], dtype=np.int8).tobytes(),
              dtype=np.uint8,
          ).flatten(),
      ),
  )
  def test_add_new_constant_tensor(
      self,
      tensor_data,
      tensor_type,
      expected_type,
      expected_shape,
      expected_buffer_data,
  ):
    """Tests if the constant tensor is added to the model."""
    ret = transformation_utils.add_new_constant_tensor(
        tensor_name="test_tensor",
        data=tensor_data,
        tensor_type=tensor_type,
        subgraph=self.model.subgraphs[0],
        buffers=self.model.buffers,
    )
    self.assertEqual(ret, len(self.model.subgraphs[0].tensors) - 1)
    self.assertEqual(
        str(self.model.subgraphs[0].tensors[-1].name), "test_tensor"
    )
    self.assertEqual(
        expected_type,
        self.model.subgraphs[0].tensors[-1].type,
    )
    self.assertEqual(
        expected_shape,
        self.model.subgraphs[0].tensors[-1].shape,
    )
    self.assertListEqual(
        expected_buffer_data.tolist(),
        self.model.buffers[
            self.model.subgraphs[0].tensors[-1].buffer
        ].data.tolist(),
    )

  @parameterized.named_parameters(
      dict(
          testcase_name="float32",
          tensor_type=schema_py_generated.TensorType.FLOAT32,
          tensor_shape=[1, 1, 1, 1],
          expected_shape=[1, 1, 1, 1],
          expected_type=schema_py_generated.TensorType.FLOAT32,
      ),
      dict(
          testcase_name="int8",
          tensor_type=schema_py_generated.TensorType.INT8,
          tensor_shape=[1, 224, 224, 1],
          expected_shape=[1, 224, 224, 1],
          expected_type=schema_py_generated.TensorType.INT8,
      ),
  )
  def test_add_new_activation_tensor_to_subgraph(
      self,
      tensor_type,
      tensor_shape,
      expected_shape,
      expected_type,
  ):
    """Tests if the activation tensor is added to the subgraph."""
    ret = transformation_utils.add_new_activation_tensor(
        tensor_name="test_tensor",
        shape=tensor_shape,
        tensor_type=tensor_type,
        subgraph=self.model.subgraphs[0],
    )
    self.assertEqual(ret, len(self.model.subgraphs[0].tensors) - 1)
    self.assertEqual(
        str(self.model.subgraphs[0].tensors[-1].name), "test_tensor"
    )
    self.assertEqual(
        expected_type,
        self.model.subgraphs[0].tensors[-1].type,
    )
    self.assertEqual(
        expected_shape,
        self.model.subgraphs[0].tensors[-1].shape,
    )

  def test_add_new_activation_tensor_with_dynamic_shape(self):
    """Tests adding an activation tensor with dynamic shape."""
    subgraph = self.model.subgraphs[0]
    new_id = transformation_utils.add_new_activation_tensor(
        tensor_name="test_tensor",
        shape=[1, -1, -1, 1],
        tensor_type=schema_py_generated.TensorType.FLOAT32,
        subgraph=subgraph,
    )
    # Originally had 4 tensors, new tensor is added at index 4.
    self.assertEqual(new_id, 4)
    self.assertLen(subgraph.tensors, 5)
    self.assertEqual(subgraph.tensors[-1].name, "test_tensor")
    self.assertEqual(
        subgraph.tensors[-1].type, schema_py_generated.TensorType.FLOAT32
    )
    self.assertEqual(subgraph.tensors[-1].shape, [1, 1, 1, 1])
    self.assertEqual(subgraph.tensors[-1].shapeSignature, [1, -1, -1, 1])


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Duplicate tensor transformation."""

from ai_edge_quantizer import qtyping
from ai_edge_quantizer.transformations import transformation_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils


def duplicate_tensor(
    transformation_input: transformation_utils.TransformationInput,
) -> qtyping.TransformationInfo:
  """Duplicates the tensor."""
  tensor_id = transformation_input.tensor_id
  subgraph = transformation_input.subgraph
  tensor = subgraph.tensors[tensor_id]
  tensor_name = tfl_flatbuffer_utils.get_tensor_name(tensor)
  buffer_data = transformation_input.buffers[tensor.buffer].data
  if buffer_data is None:
    raise ValueError(
        'Duplicate Tensor transformation supports only constant tensors.'
        f' Tensor {tensor_name} is not constant.'
    )
  new_tensor_id = transformation_utils.add_new_constant_tensor(
      tensor_name=f'{tensor_name}_duplicated',
      data=buffer_data,
      tensor_type=tensor.type,
      tensor_shape=tensor.shape,
      subgraph=subgraph,
      buffers=transformation_input.buffers,
      force_duplicate_buffer=True,
  )
  # Update the tensor name to avoid name collision in case when tensor is
  # duplicated mulitple times.
  subgraph.tensors[new_tensor_id].name += f'_{new_tensor_id}'

  # Update the consumers' input tensor id to the duplicated tensor id.
  # Assuming transformation_input to contain all and only consumers that are
  # supposed to use this new duplicated tensor.
  for consumer in transformation_input.consumers:
    consumer_inputs = subgraph.operators[consumer].inputs
    for i in range(len(consumer_inputs)):
      if consumer_inputs[i] == tensor_id:
        consumer_inputs[i] = new_tensor_id
        break

  return qtyping.TransformationInfo(
      op_id=0, num_ops_added=0, output_tensor_id=new_tensor_id
  )
